{ **************************************************************************** }
{  OpenAL OGG-Streaming - Port  2003 by Sascha Willems                        }
{ **************************************************************************** }
{  A conversion of devmaster.net's OpenAL-Tutorial no.8                        }
{  (http://www.devmaster.net/articles/openal-tutorials/lesson8.php)            }
{                                                                              }
{   visit http://www.delphigl.de for more                                      }
{   visit http://www.delphigl.com the biggest german Delphi+OpenGL community   }
{                                                                              }
{  More infos on the OpenAL-SoundAPI can be found here :                       }
{   http://oss.sgi.com/projects/ogl-sample/registry/ARB/vertex_buffer_object.t }
{                                                                              }
{  Additional infos on the ogg-vorbis fileformat can be obtained here :        }
{   http://www.vorbis.com                                                      }
{                                                                              }
{  Uses the follwowing libs :                                                  }
{   OGG-Lib for Delphi by Matthijs Laan (http://www.delphi-jedi.org)           }
{   OpenAL-Lib for Delphi by Tom Nuydens (http://www.delphi3d.net)             }
{                                                                              }
{ **************************************************************************** }
{  Compiled with Delphi 7.0                                                    }
{ *****************************************************************************}
{  Please let me know if you use some of this in your project and don't foret: }
{  Use at your own risk ;)                                                     }
{ *****************************************************************************}

unit OpenAL_OGG_Form;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  ExtCtrls,
  StdCtrls,

  // OpenAL units
  al,
  altypes,
  alut,

  // OGG-Vorbis units
  ogg,
  codec,
  vorbisenc,
  vorbisfile;

const
 BufferSize = 4096*8;

type
  TALForm = class(TForm)
    Label1: TLabel;
    Memo1: TMemo;
    Button1: TButton;
    LogTimer: TTimer;
    Button2: TButton;
    OGGTimer: TTimer;
    SeekPos: TLabel;
    Button3: TButton;
    OpenDialog: TOpenDialog;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure LogTimerTimer(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure OGGTimerTimer(Sender: TObject);
    procedure Button3Click(Sender: TObject);
   private
   public
   end;

  TOGGStream = class(TObject)
    public
     Log : TStringList;
     constructor Create;
     destructor Destroy; override;
     procedure Open(pFileName : String);
     procedure Release;
     procedure Display;
     function Playback : Boolean;
     function Playing : Boolean;
     function Update : Boolean;
    protected
     function Stream(Buffer : TALUInt) : Boolean;
     function ErrorString(Code : Integer) : String;
     procedure Empty;
     procedure Check;
    private
     OGGFile       : TFileStream;
     OGGStream     : OGGVorbis_File;
     VorbisInfo    : Vorbis_Info;
     VorbisComment : Vorbis_Comment;
     Buffers       : array[0..1] of TALUInt;
     Source        : TALUInt;
     Format        : TALEnum;
   end;

var
  ALForm    : TALForm;
  OGGStream : TOGGStream;
  OGGName   : String;

implementation

{$R *.dfm}

// =============================================================================
//  TOGGStream
// =============================================================================

// =============================================================================
//  TOGGStream.Create                                            - CONSTRUCTOR -
// =============================================================================
constructor TOGGStream.Create;
begin
inherited;
Log := TStringList.Create;
end;

// =============================================================================
//  TOGGStream.Destroy                                            - DESTRUCTOR -
// =============================================================================
destructor TOGGStream.Destroy;
begin
Log.SaveToFile('openAL_log.txt');
Log.Free;
OggFile.Free;
inherited;
end;

// =============================================================================
//  TOGGStream.Open
// =============================================================================
procedure TOggStream.Open(pFileName : String);
var
 Res : Integer;
begin
if Assigned(OggFile) then
 OggFile.Free;
if not FileExists(pFileName) then
 begin
 Log.Add('Could not open Ogg file');
 ShowMessage('File not found!');
 Halt(0);
 end;
// Load the OGG-File into a Filestream
OggFile := TFileStream.Create(pFileName, fmOpenRead);
// we can't use ov_open in Delphi,as this function works with C-filehandles.So
// we use ov_open_callbacks instead which uses a filestream
Res := ov_open_callbacks(OggFile, OggStream, nil, 0, ops_callbacks);
if Res <> 0 then
 begin
 Log.Add('Could not open Ogg stream. ['+ErrorString(Res)+']');
 exit;
 end;
// Get some infos out of the OGG-file
VorbisInfo    := ov_info(OggStream,-1)^;
VorbisComment := ov_comment(OggStream,-1)^;
if VorbisInfo.channels = 1 then
 Format := AL_FORMAT_MONO16
else
 Format := AL_FORMAT_STEREO16;
// Prepare OpenAL for OGG-Streaming (i.e. generate buffers and sources)
alGenBuffers(2, @Buffers);
Check;
alGenSources(1, @Source);
Check;
alSource3f(source, AL_POSITION, 0, 0, 0);
alSource3f(source, AL_VELOCITY, 0, 0, 0);
alSource3f(source, AL_DIRECTION, 0, 0, 0);
alSourcef(source, AL_ROLLOFF_FACTOR, 0);
alSourcei(source, AL_SOURCE_RELATIVE, AL_TRUE);
end;

// =============================================================================
//  TOGGStream.Release
// =============================================================================
procedure TOggStream.Release;
begin
alSourceStop(source);
Empty;
alDeleteSources(1, @source);
Check;
alDeleteBuffers(1, @Buffers);
Check;
ov_clear(OggStream);
end;

// =============================================================================
//  TOGGStream.Display
// =============================================================================
procedure TOggStream.Display;
begin
with VorbisInfo do
 begin
 Log.Add('Version         = '+IntToStr(Version));
 Log.Add('Channels        = '+IntToStr(Channels));
 Log.Add('Rate (hz)       = '+IntToStr(Rate));
 Log.Add('Bitrate upper   = '+IntToStr(bitrate_upper));
 Log.Add('Bitrate nominal = '+IntToStr(bitrate_nominal));
 Log.Add('Bitrate lower   = '+IntToStr(bitrate_lower));
 Log.Add('Bitrate window  = '+IntToStr(bitrate_window));
 Log.Add('Vendor          = '+VorbisComment.Vendor);
 end;
end;

// =============================================================================
//  TOGGStream.Playback
// =============================================================================
function TOggStream.Playback : Boolean;
begin
if Playing then
 begin
 Result := True;
 exit;
 end;
if not Stream(Buffers[0]) then
 begin
 Result := False;
 exit;
 end;
if not Stream(Buffers[1]) then
 begin
 Result := False;
 exit;
 end;
alSourceQueueBuffers(Source, 2, @Buffers);
alSourcePlay(Source);
Result := True;
end;

// =============================================================================
//  TOGGStream.Playing
// =============================================================================
function TOggStream.Playing : Boolean;
var
 State : TALEnum;
begin
alGetSourcei(Source, AL_SOURCE_STATE, @State);
Result := (State = AL_PLAYING);
end;

// =============================================================================
//  TOGGStream.Update
// =============================================================================
function TOggStream.Update : Boolean;
var
 Processed : Integer;
 Active    : Boolean;
 Buffer    : TALUInt;
begin
Active := False;
if OGGStream.offset >= OGGStream.end_v then
 begin
 Result := False;
 exit;
 end;
alGetSourcei(Source, AL_BUFFERS_PROCESSED, @Processed);
if Processed > 0 then
 repeat
 alSourceUnqueueBuffers(source, 1, @Buffer);
 Check;
 Active := Stream(Buffer);
 alSourceQueueBuffers(source, 1, @Buffer);
 Check;
 dec(Processed);
 until Processed <= 0;
Result := Active;
end;

// =============================================================================
//  TOGGStream.Stream
// =============================================================================
function TOggStream.Stream(Buffer : TALUInt) : Boolean;
var
 Data    : PChar;
 Size    : Integer;
 Section : Integer;
 Res     : Integer;
begin
Size := 0;
GetMem(Data, BufferSize+SizeOf(Size));
while (Size < BufferSize) do
 begin
 Res := ov_read(OggStream,PByteArray(Data)^[size],BufferSize-Size,0,2,1,@Section);
 if Res > 0 then
  inc(Size,Res)
 else
  if Res < 0 then
   Log.Add(ErrorString(Res))
  else
   break;
 end;
if Size = 0 then
 begin
 Result := False;
 exit;
 end;
alBufferData(Buffer, Format, Data, Size, VorbisInfo.Rate);
Check;
Result := True;
FreeMem(Data);
end;

// =============================================================================
//  TOGGStream.Empty
// =============================================================================
procedure TOGGStream.Empty;
var
 Queued : Integer;
 Buffer : TALUInt;
 i      : Integer;
begin
alGetSourcei(source, AL_BUFFERS_QUEUED, @Queued);
for i := 0 to Queued-1 do
 begin
 alSourceUnqueueBuffers(Source, 1, @Buffer);
 Check;
 end;
end;

// =============================================================================
//  TOGGStream.Check
// =============================================================================
procedure TOGGStream.Check;
var
 Error : Integer;
begin
Error := alGetError;
if Error <> AL_NO_ERROR then
 Log.Add('OpenAL error was raised.');
end;

// =============================================================================
//  TOGGStream.ErrorString
// =============================================================================
function TOGGStream.ErrorString(Code : Integer) : String;
begin
case Code of
 OV_EREAD      : Result := 'Read from Media.';
 OV_ENOTVORBIS : Result := 'Not Vorbis data.';
 OV_EVERSION   : Result := 'Vorbis version mismatch.';
 OV_EBADHEADER : Result := 'Invalid Vorbis header.';
 OV_EFAULT     : Result := 'nternal logic fault (bug or heap/stack corruption.';
else
 Result := 'Unknown Ogg error.';
end;
end;

// =============================================================================
//  TALForm                                                         - MAINFORM -
// =============================================================================
procedure TALForm.FormCreate(Sender: TObject);
var
 ArgV : array of PChar;
begin
// Init OpenAL
SetLength(ArgV,0);
alutInit(nil,ArgV);
// Create OGGStream object
OGGStream := TOGGStream.Create;
end;

procedure TALForm.FormDestroy(Sender: TObject);
begin
alutExit;
OGGStream.Free;
end;

procedure TALForm.LogTimerTimer(Sender: TObject);
var
 i : Integer;
begin
Memo1.Lines.Clear;
if OGGStream.Log.Count > 0 then
 for i := 0 to OGGStream.Log.Count-1 do
  Memo1.Lines.Add(OGGStream.Log.Strings[i]);
end;

procedure TALForm.Button1Click(Sender: TObject);
begin
Close;
end;

procedure TALForm.Button2Click(Sender: TObject);
begin
if OGGName = '' then
 begin
 ShowMessage('No file selected!');
 exit;
 end;
OGGStream.Log.Clear;
OGGTimer.Enabled := False;
// Load OGG-File and play it
with OGGStream do
 begin
 // Change this to the filename you want to play...
 Open(OGGName);
 Display;
 if not Playback then
  begin
  Log.Add('Ogg refused to play.');
  exit;
  end;
 end;
OGGTimer.Enabled := True;
end;

procedure TALForm.OGGTimerTimer(Sender: TObject);
begin
SeekPos.Caption := IntToStr(OggStream.OGGStream.offset)+'/'+IntToStr(OggStream.OGGStream.end_v);
with OGGStream do
 begin
 Update;
 if not Playing then
  if not Playback then
   begin
   Log.Add('Ogg abruptly stopped.');
   exit;
   end
  else
   Log.Add('Ogg stream was interrupted.');
 end;
end;

procedure TALForm.Button3Click(Sender: TObject);
begin
if OpenDialog.Execute then
 begin
 OGGName := OpenDialog.FileName;
 Button2Click(self);
 end;
end;

end.
